;;; $Id: type-check-type-translate.scm,v 1.6 1999/06/23 16:57:22 leavens Exp $
;;; Translation from the internal type format to the external format,
;;; with nested arguments, and better variable names for types.
;;; AUTHORS: Curtis Clifton and Gary T. Leavens

(load-quietly-from-lib "type-check-external-type-expr.scm")

(define tc:type-translate
  ;; TYPE: (-> (type-expr) external-type-expr)
  (lambda (typ)
    ;; ENSURES: result is typ with argument types nested, and with
    ;;          type variables translated into letters if possible.
    (cond
     ((tc:variable-type-expr? typ)
      (tc:make-variable-external-type-expr 'T))
     (else (tc:nest-args (tc:better-variables typ (tc:type-vars typ)))))))

(define tc:union
  ;; TYPE: (-> ((list symbol) (list symbol)) (list symbol))
  (lambda (s1 s2)
    ;; REQUIRES: s2 has no duplicates
    ;; ENSURES: result has the elements of s1 and s2,
    ;;         but no duplicates
    (cond
     ((null? s1) s2)
     ((memq (car s1) s2) (tc:union (cdr s1) s2))
     (else (cons (car s1) (tc:union (cdr s1) s2))))))

(define tc:unionAll
  ;; TYPE: (-> ((list (list symbol))) (list symbol))
  (lambda (ss)
    ;; REQUIRES: each element of ss has no duplicates
    ;; ENSURES: result has the elements of the elements of ss,
    ;;         but no duplicates
    (if (null? ss)
	'()
	(tc:union (car ss) (tc:unionAll (cdr ss))))))

(define tc:type-vars
  ;; TYPE: (-> (type-expr) (list symbol))
  (lambda (typ)
    ;; ENSURES: result is a list of all the type variables in typ,
    ;;          and no element of result occurs twice in result
    (cond
     ((tc:variable-type-expr? typ)
      (list (tc:variable-type-expr->name typ)))
     ((tc:non-variable-basic-type-expr? typ) '())
     ((tc:function-type-expr? typ)
      (tc:union
       (tc:unionAll (map tc:type-vars
			 (tc:function-type-expr->arg-types typ)))
       (tc:type-vars
	(tc:function-type-expr->result-type typ))))
     ((tc:and-type-expr? typ)
      (tc:unionAll (map tc:type-vars
			(tc:and-type-expr->conjoined-types typ))))
     ((tc:applied-type-expr? typ)
      (tc:union (tc:type-vars (tc:applied-type-expr->operator-type typ))
		(tc:unionAll (map tc:type-vars
				  (tc:applied-type-expr->operand-types typ)))))
     (else
      (error
       "tc:type-vars called with unknown type expression:"
       typ))
     )))

(define tc:better-variables
  ;; TYPE: (-> (type-expr (list symbol)) type-expr)
  (lambda (typ vars)
    ;; ENSURES: result is typ with some type variables in typ
    ;;          replaced by better variable names
    (letrec
	((helper
	  ;; TYPE: (-> (type-expr (list symbol) (list symbol))
	  ;;           type-expr)
	  (lambda (typ vars better-names)
	    ;; ENSURES: result is typ with each occurrence of vars replaced
	    ;;          with a better-name, until the better names run out...
	    (cond
	     ((or (null? vars) (null? better-names))
	      typ)
	     (else
	      (helper (tc:substq-all (car better-names) (car vars) typ)
		      (cdr vars)
		      (cdr better-names)))))))
      (helper typ
	      vars
	      (if (= 1 (length vars))
		  '(T)
		  '(S T U V W X Y Z
		    A B C D E F G H I J K L M N O P Q R))))))

(define tc:substq-all
  ;; TYPE: (-> (symbol symbol type-expr) type-expr)
  (lambda (new old typ)
    ;; ENSURES: result is typ with all occurrences of old replaced by new
    (cond
     ((tc:variable-type-expr? typ)
      (if (eq? old (tc:variable-type-expr->name typ))
	  (tc:make-variable-type-expr new)
	  typ))
     ((tc:non-variable-basic-type-expr? typ) typ)
     ((tc:function-type-expr? typ)
      (tc:make-function-type-expr
       (map (lambda (t) (tc:substq-all new old t)) 
	    (tc:function-type-expr->arg-types typ))
       (tc:substq-all
	new
	old
	(tc:function-type-expr->result-type typ))))
     ((tc:and-type-expr? typ)
      (tc:make-and-type-expr
       (map (lambda (t) (tc:substq-all new old t))
	    (tc:and-type-expr->conjoined-types typ))))
     ((tc:applied-type-expr? typ)
      (tc:make-applied-type-expr
       (tc:substq-all new old (tc:applied-type-expr->operator-type typ))
       (map (lambda (t) (tc:substq-all new old t))
	    (tc:applied-type-expr->operand-types typ))))
     (else
      (error
       "tc:substq-all called with unknown type expression:"
       typ))
     )))

(define tc:nest-args
  ;; TYPE: (-> (type-expr) external-type-expr)
  (lambda (typ)
    ;; ENSURES: result is an external-type-expr equivalent to typ
    (cond
     ((tc:variable-type-expr? typ)
      (tc:make-variable-external-type-expr
       (tc:variable-type-expr->name typ)))
     ((tc:non-variable-basic-type-expr? typ)
      (tc:make-basic-external-type-expr (tc:basic-type-expr->symbol typ)))
     ((tc:function-type-expr? typ)
      (tc:make-function-external-type-expr
       (map tc:nest-args (tc:function-type-expr->arg-types typ))
       (tc:nest-args (tc:function-type-expr->result-type typ))))
     ((tc:and-type-expr? typ)
      (tc:make-and-external-type-expr
       (map tc:nest-args (tc:and-type-expr->conjoined-types typ))))
     ((tc:applied-type-expr? typ)
      (tc:make-applied-external-type-expr
       (tc:external-type-expr->basic-type
	(tc:nest-args (tc:applied-type-expr->operator-type typ)))
       (map tc:nest-args (tc:applied-type-expr->operand-types typ))))
     (else (error "tc:nest-args called with unknown type expression:" typ))
     )))

(define tc:type-untranslate
  ;; TYPE: (-> (external-type-expr) type-expr)
  (lambda (typ)
    ;; ENSURES: result is typ with argument types nested, and with
    ;;          single letter names (external type variables)
    ;;          translated into type variables.
    (let ((return-val
	   (tc:unnest-args (tc:worse-variables typ (tc:external-type-vars typ)))))
      return-val)))

(define tc:external-type-vars
  ;; TYPE: (-> (external-type-expr) (list symbol))
  (lambda (typ)
    ;; ENSURES: result is a list of all the type variables in typ,
    ;;          and no element of result occurs twice in result
    (cond
     ((tc:external-type-expr-variable? typ)
      (list (tc:variable->symbol (tc:external-type-expr->variable typ))))
     ((tc:external-type-expr-basic-type? typ) '())
     ((tc:external-type-expr-function-type? typ)
      (tc:union
       (tc:unionAll (map tc:external-type-vars
			 (tc:external-type-expr->function-arg-types typ)))
       (tc:external-type-vars
	(tc:external-type-expr->function-result-type typ))))
     ((tc:external-type-expr-and-type? typ)
      (tc:unionAll (map tc:external-type-vars
			(tc:external-type-expr->conjoined-types typ))))
     ((tc:external-type-expr-applied-type? typ)
      (tc:union (tc:external-type-vars
		 (tc:make-basic-external-type-expr
		  (tc:external-type-expr->operator typ)))
		(tc:unionAll (map tc:external-type-vars
				  (tc:external-type-expr->operands typ)))))
     (else
      (error
       "tc:external-type-vars called with unknown external type expression:"
       typ))
     )))

(define tc:worse-variables
  ;; TYPE: (-> (external-type-expr (list symbol)) external-type-expr)
  (lambda (typ vars)
    ;; ENSURES: result is typ with some type variables in typ
    ;;          replaced by "worse" variable names
    (letrec
	((helper
	  ;; TYPE: (-> (external-type-expr (list symbol) (list symbol))
	  ;;           external-type-expr)
	  (lambda (typ vars worse-names)
	    ;; ENSURES: result is typ with each occurrence of a symbol in vars
	    ;;          that is not a tc:variable? replaced
	    ;;          with a worse-name
	    (cond
	     ((or (null? vars) (null? worse-names))
	      typ)
	     ((tc:variable? (car vars))
	      (helper typ (cdr vars) worse-names))
	     (else
	      (helper (tc:substq-all-external (car worse-names) (car vars) typ)
		      (cdr vars)
		      (cdr worse-names)))))))
      (helper typ
	      vars
	      (if (= 1 (length vars))
		  '(?Twn)
		  '(?Swn ?Twn ?Uwn ?Vwn ?Wwn ?Xwn ?Ywn ?Zwn
			 ?Awn ?Bwn ?Cwn ?Dwn ?Ewn ?Fwn ?Gwn ?Hwn ?Iwn
			 ?Jwn ?Kwn ?Lwn ?Mwn ?Nwn ?Own ?Pwn ?Qwn ?Rwn))))))

(define tc:substq-all-external
  ;; TYPE: (-> (symbol symbol external-type-expr) external-type-expr)
  (lambda (new old typ)
    ;; ENSURES: result is tr with all occurrences of old replaced by new
    (cond
     ((tc:external-type-expr-variable? typ)
      (if (eq? old (tc:external-type-expr->variable typ))
	  (tc:make-variable-external-type-expr new)
	  typ))
     ((tc:external-type-expr-basic-type? typ) typ)
     ((tc:external-type-expr-function-type? typ)
      (tc:make-function-external-type-expr
       (map (lambda (t) (tc:substq-all-external new old t)) 
	    (tc:external-type-expr->function-arg-types typ))
       (tc:substq-all-external
	new
	old
	(tc:external-type-expr->function-result-type typ))))
     ((tc:external-type-expr-and-type? typ)
      (tc:make-and-external-type-expr
       (map (lambda (t) (tc:substq-all-external new old t))
	    (tc:external-type-expr->conjoined-types typ))))
     ((tc:external-type-expr-applied-type? typ)
      (tc:make-applied-external-type-expr
       (tc:external-type-expr->basic-type
	(tc:substq-all-external new old
				(tc:make-basic-external-type-expr
				 (tc:external-type-expr->operator typ))))
       (map (lambda (t) (tc:substq-all-external new old t))
	    (tc:external-type-expr->operands typ))))
     (else
      (error
       "tc:substq-all-external called with unknown type expression:"
       typ))
     )))

(define tc:unnest-args
  ;; TYPE: (-> (external-type-expr) type-expr)
  (lambda (typ)
    ;; ENSURES: result is a type-expr equivalent to typ
    (cond
     ((tc:external-type-expr-variable? typ)
      (tc:make-variable-type-expr
       (tc:variable->symbol (tc:external-type-expr->variable typ))))
     ((tc:external-type-expr-basic-type? typ)
      (tc:make-basic-type-expr (tc:external-type-expr->basic-type typ)))
     ((tc:external-type-expr-function-type? typ)
      (tc:make-function-type-expr
       (map tc:unnest-args (tc:external-type-expr->function-arg-types typ))
       (tc:unnest-args (tc:external-type-expr->function-result-type typ))))
     ((tc:external-type-expr-and-type? typ)
      (tc:make-and-type-expr
       (map tc:unnest-args (tc:external-type-expr->conjoined-types typ))))
     ((tc:external-type-expr-applied-type? typ)
      (tc:make-applied-type-expr
	(tc:unnest-args (tc:make-basic-external-type-expr
			 (tc:external-type-expr->operator typ)))
       (map tc:unnest-args (tc:external-type-expr->operands typ))))
     (else (error "tc:unnest-args called with unknown type expression:" typ))
     )))
