;;; $Id: type-check-utils.scm,v 1.12 1999/06/23 16:57:22 leavens Exp $
;;; SYNOPSIS: This file contains some general utility procedures that are 
;;;           used by procedures in the tc:type checker.
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

(load-quietly-from-lib "displayln.scm")
(load-quietly-from-lib "atomic-tree.scm")

(define tc:small-error
  ;; TYPE: (-> (datum ...) void)
  ;; EFFECT: displays an error message w/o resetting the interpreter
  ;;         This is used to do recoverable errors w/o  using
  ;;         call/cc
  (lambda args
    (display "Error: ")
    (apply displayln args)))

(define tc:but-last
  ;; TYPE: (-> ((list T)) (list T))
  (lambda (lst)  
    ;; REQUIRES: lst is not empty
    ;; EFFECT: returns all but the last element of lst
    (if (null? (cdr lst))
	'()
	(cons (car lst)
	      (tc:but-last (cdr lst))))))

(define tc:last-item ;; TYPE: (-> ((list T)) T)
  (lambda (ls)
    ;; REQUIRES: ls is not empty
    ;; ENSURES: result is the last item of ls
    (cond
      ((null? (cdr ls)) (car ls))
      (else (tc:last-item (cdr ls))))))

(define tc:deep-recur-abs
  ;; TYPE: (-> (T (-> (S T) T) (-> (T T) T) (-> (datum) boolean))
  ;;           (-> ((tree S)) T))
  (lambda (seed flat-proc tree-proc leaf?)
    ;; REQUIRES: (leaf? x) = (x has tc:type S)
    ;; ENSURES: (result '()) = seed,
    ;;          (result (cons x t))  = (flat-proc x (result t)),
    ;;          (result (cons t1 t2)) = (tree-proc (result t1) (result t2))
    ;;          where (leaf? x) holds, but not (leaf? t1) and (leaf? t2).
    (letrec
      ((helper  ;; TYPE: (-> ((tree S)) T)
         (lambda (tr)
           (if (null? tr)
               seed
               (let ((a (car tr)))
                 (if (leaf? a)
                     (flat-proc a (helper (cdr tr)))
                     (tree-proc (helper a) (helper (cdr tr)))))))))
      helper)))


;;; The following allows easy creation of deep recursions schemes over
;;; "trees of S", by using a leaf? predicate that tests for type S.

(define tc:deep-recur-abs-m
  ;; TYPE: (-> ((-> (datum) boolean))
  ;;           (-> (T (-> (S T) T) (-> (T T) T) (-> ((tree S)) T))))
  (lambda (leaf?)
    (lambda (seed flat-proc tree-proc)
      (tc:deep-recur-abs seed flat-proc tree-proc leaf?))))


;;; The following is the case of deep recursion over trees of atomic-items.
;;; Compare with program 7.28 in Scheme and the Art of Programming

(define tc:deep-recur
  ;; TYPE: (-> (T (-> (S T) T) (-> (T T) T)) (-> ((tree S)) T))
  (let ((atomic-item?
	 ;; TYPE: (-> (datum) boolean)
	 (lambda (x)
	   (not (or (pair? x)
		    (null? x))))))
    ;; REQUIRES: if x has tc:type S, then (atomic-item? x) is true
    (tc:deep-recur-abs-m atomic-item?)))


(define tc:replace
  ;;  TYPE: (-> (T T (tree T)) (tree T))
  (lambda (old new tree)
    (cond
     ((tree-null? tree) tree)
     ((tree-first-tree? tree)
      (tree-add-tree (tc:replace old new (tree-first-tree tree))
		     (tc:replace old new (tree-rest tree))))
     ((eq? old (tree-first-atom tree))
      (tree-add-atom new (tc:replace old new (tree-rest tree))))
     (else (tree-add-atom (tree-first-atom tree)
			  (tc:replace old new (tree-rest tree)))))))


(define tc:singleton-list?
  ;;  TYPE: (-> (datum) boolean)
  (lambda (xp)
    (and (pair? xp) (null? (cdr xp)))))


(define tc:flat-recur
  ;; TYPE: (-> (T (-> (S T) T)) (-> ((list S)) T))
  (lambda (seed list-proc)
    (letrec
	((helper
	  (lambda (ls)
	    (if (null? ls)
		seed
		(list-proc (car ls) (helper (cdr ls)))))))
      helper)))  

(define tc:first car)   ;; TYPE: (-> ((list T)) T)
(define tc:second cadr) ;; TYPE: (-> ((list T)) T)
(define tc:third caddr) ;; TYPE: (-> ((list T)) T)

(define tc:map2
  ;; TYPE: (-> ((-> (S T) U) (list S) (list T)) (list U))
  (lambda (proc ls1 ls2)
    ;; REQUIRES: ls1 and ls2 have same length.
    ;; ENSURES: the result is a list, its items come from the result of
    ;;          proc operating on corresponding items in ls1 and ls2.
    (if (null? ls1) '()
	(cons (proc (car ls1) (car ls2))
	      (tc:map2 proc (cdr ls1) (cdr ls2))))))

(define tc:all
  ;; TYPE: (-> ((-> (T) boolean) (list T)) boolean)
  (lambda (pred ls)
    ;; ENSURES: result is true just when pred is true for each item in ls.
    (or (null? ls)
	(and (pred (car ls))
	     (tc:all pred (cdr ls))))))

(define tc:some
  ;; TYPE: (-> ((-> (T) boolean) (list T)) boolean)
  (lambda (pred ls)
    ;; ENSURES: result is true when pred is true for any item in ls.
    (and (not (null? ls))
	 (or (pred (car ls))
	     (tc:some pred (cdr ls))))))

(define tc:member?
  ;; TYPE: (-> (T (list T)) boolean)
  (lambda (item ls)
    (and (not (null? ls))
	 (or (equal? item (car ls))
	     (tc:member? item (cdr ls))))))

(define tc:scheme-exp-member?
  ;; TYPE: (-> (datum scheme-exp) boolean)
  (lambda (item tr)
    (cond
     ((not (pair? tr)) (equal? item tr))
     ((null? tr) #f)
     (else
      (or (tc:scheme-exp-member? item (car tr))
	  (tc:scheme-exp-member? item (cdr tr)))))))

(define tc:vector-for-each
  ;; TYPE: (-> ((-> (T) void) (vector T) number) void)
  (lambda (f v start)
    ;; EFFECT: apply f to each element of v, starting at index start
    ;;         and going towards larger indexes
    (let ((size (vector-length v)))
      (letrec ((helper
		;; TYPE: integer -> void
		(lambda (i)
		  (if (not (= i size))
		      (begin
			(f (vector-ref v i))
			(helper (+ 1 i)))))))
	(helper start)))))

(define tc:ignore
  ;; TYPE: (-> (T) void)
  (lambda (x) (set! x x)))
