;;; $Id: type-check.scm,v 1.47 1999/06/23 16:57:22 leavens Exp $
;;; Polymorphic type checking for Scheme
;;;
;;; SYNOPSIS: the main useful routines are:
;;;           type-check-file, which reads expressions from a file,
;;;                            and checks them.
;;;           type-check-loop, which reads expressions or definitions and type
;;;                            checks them
;;;           type-check, which takes a single (quoted) expression
;;;                       and type checks it
;;;
;;; BUGS: Procedures etc. must be defined before being used.
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

(load-quietly-from-lib "type-check-polytype.scm")
(load-quietly-from-lib "type-check-down-type.scm")

;;; Set the following flags to the appropriate values for your needs.
;;; Setting any of these to #f will warn about constructs that you may not
;;; want to allow users to have, or that may be dangerous.
;;; In particular, if you are doing SICP style block-structure,
;;; you'll need to set *allow-nested-defines* to #t.

(define *allow-nested-defines* #t)
(define *allow-cond-expressions-without-else-clause* #t)
(define *allow-case-expressions-without-else-clause* #t)
(define *warn-about-implicit-begins* #f)

(define tc:implicit->explicit
  ;; TYPE: (-> (string)
  ;;           (-> ((list s-expression)) s-expression))
  (lambda (context-name)
    (lambda (exps)
      (if (null? (cdr exps))
	  (tc:desugar (car exps))
	  (begin
	    (if *warn-about-implicit-begins*
		(displayln "Warning: implicit begin in " context-name ": "
			   exps))
	    (cons 'begin (map tc:desugar exps)))))))

(define tc:type-check-maker
  ;; TYPE: (-> ((-> (symbol) void) (-> (external-type-expr) T))
  ;;           (-> (s-expression) T))
  (lambda (display-defined-name type-return)
    ;; Gary: I changed the type of type-return to take an
    ;; external-type-expr.
    ;;                                      - Curt 5/26/99
    ;;
    (let ((new-global;; TYPE: (-> (symbol type-expr) T)
	   (lambda (name its-type)
	     (set! *tc:global-var-types*
		   (tc:env-update name its-type *tc:global-var-types*))
	     (display-defined-name name)
	     (type-return (tc:type-translate its-type)))))
      (lambda (expr)
	;; MODIFIES: *tc:global-var-types*
	;; EFFECT: if expr has a type, return the result of passing it
	;;         to the procedure type-return, otherwise give an error.
	;;         For define expressions, installs the name and its type
	;;         in *tc:global-var-types*.
	(if (>= *tc:tracing-level* 2)
	    (begin
	      (displayln "tc:type-check-maker checking")
	      (pretty-print expr)))
	(cond
	 ((or (procedure? expr) (vector? expr))
	  (begin 
	    (error "you have to quote your expression argument: " expr)
	    (type-return *tc:external-failure*))) ; we don't get here but this
					; helps the type checker infer
					; the type of type-return
	 ((not (pair? expr))
	  (type-return (tc:type-translate (tc:type expr))))
	 ((tc:define-exp? expr)
	  (let ((exp (tc:sicp->sap expr)))
	    (if (and (tc:scheme-exp-member? 'define (cdr exp)) 
		     (not *allow-nested-defines*))
		(error "use of nested define in: " expr))
	    (new-global (cadr exp)
			(tc:type (tc:desugar
				  (tc:nested-define->letrec exp))))))
	 ((tc:grouped-defines? expr)
	  (for-each type-check (cdr expr))
	  (type-return (tc:make-basic-external-type-expr 'void)))
	 ((tc:grouped-deftypes? expr)
	  (for-each type-check (cdr expr))
	  (type-return (tc:make-basic-external-type-expr 'void)))
	 ((tc:deftype? expr)
	  (begin
	    ;; (displayln "deftype: " expr)
	    (if (= (length expr) 3)
		(new-global (cadr expr) (tc:type-untranslate (caddr expr)))
		(error "bad syntax for deftype: " expr))))
	 ((tc:define-record? expr)
	  (if (not (and (= (length expr) 3)
			(symbol? (cadr expr))))
	      (error "bad syntax for define-record: " expr)
	      ((tc:type-check-maker display-defined-name type-return)
	       (tc:define-record->deftypes expr))))
	 (else (type-return (tc:type-translate
			     (tc:type (tc:desugar expr))))))))))

(define tc:define-record?
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (pair? x)
	 (eq? (car x) 'define-record))))

(define tc:deftype?
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (pair? x)
	 (eq? (car x) 'deftype))))

(define tc:colon " : ") ; TYPE: string
(define tc:cannot-be-typed "cannot be typed") ; TYPE: string
  
(define type-check
  ;; TYPE: (-> (expression) void)
  (tc:type-check-maker
   (lambda (name)
     (display name)
     (display tc:colon))
   (lambda (typ)
     (tc:type-write typ)
     (newline))))

(define type-check-file 
  ;; TYPE: (-> (string) void)
  (lambda (fname)
    (tc:type-check-file fname #f)))

(define *tc:loaded-lib-files*
  ;; TYPE: (list string)
  '())

(define tc:type-check-file
  ;; TYPE: (-> (string boolean) void)
  (let ((libdir-name-length (string-length *cs342:libdir*)))
    (lambda (fname quiet)
      (cond
       ((not (string? fname))
	(error "type-check-file expects a string argument, not: " fname))
       ((tc:is-def-file? fname)
	(error "Can't type-check a *.def file: " fname))
       ((file-exists? fname)
	(if (member fname *tc:loaded-lib-files*)
	    (if (not quiet)
		(begin
		  (display "Skipping type checking.  ")
		  (displayln fname " already checked.")))
	    (begin
	      (if (and (>= (string-length fname) libdir-name-length)
		       (string-ci=?
			*cs342:libdir*
			(substring fname 0 libdir-name-length)))
		  (set! *tc:loaded-lib-files*
			(cons fname *tc:loaded-lib-files*)))
	      (cond
	       ((file-exists? (string-append (tc:get-file-prefix fname)
					     ".def"))
		(tc:handle-defs-file fname quiet))
	       (else (tc:type-check-file-aux fname quiet))))))
       ((not (tc:is-scm-file? fname))
	(tc:type-check-file (string-append fname ".scm") quiet))
       (else (error "cannot find file: " fname))))))

(define tc:announce-file
  ;; TYPE; (-> (string boolean) void)
  (lambda (fname quiet)
    (if (not quiet)
	(begin
	  (display "Type checking ")
	  (write fname)
	  (display " ...")
	  (newline)))))

(define tc:back-from-load
  ;; TYPE: (-> (string boolean) void)
  (lambda (fname quiet)
    (if (not quiet)
	(begin
	  (display "... back to type checking of ")
	  (write fname)
	  (display " ...")
	  (newline)))))

(define tc:process-loads
  ;; TYPE: (-> ((list (pair symbol string))) void)
  (lambda (load-list fname quiet)
    (for-each
     (lambda (exp)
       (let ((load-fun (car exp)))
       (cond
	((or (eq? load-fun 'load)
	     (eq? load-fun 'load-quietly))
	 (let ((this-one-is-quiet (or quiet
				      (eq? load-fun 'load-quietly))))
	   (tc:type-check-file (cadr exp) this-one-is-quiet)
	   (if (not this-one-is-quiet)
	       (tc:back-from-load fname quiet))))
	((or (eq? load-fun 'load-from-lib)
	     (eq? load-fun 'load-quietly-from-lib))
	 (let ((fullname (string-append *cs342:libdir* (cadr exp)))
	       (this-one-is-quiet (or quiet
				      (eq? load-fun 'load-quietly-from-lib))))
	   (tc:type-check-file fullname this-one-is-quiet)
	   (if (not this-one-is-quiet)
	       (tc:back-from-load fname quiet))))
	(else (error "Should not happen, missing load kind!"))
	)))
     load-list)))

(define tc:e-env
  ;; TYPE: (env uniq-sym type-expr)
  (tc:env-empty))

(define tc:type-check-file-aux
  ;; TYPE: (-> (string boolean) void)
  (lambda (fname quiet)
    ;; EFFECT: reads in a file, loads all declarations into
    ;; a single letrec declaration, then creates a list consisting of
    ;; (name . tc:type) pairs for all names declared.  Finally, it outputs
    ;; all of the names and tc:types to the user.  It also handles any needed
    ;; loads for the file.
    (tc:announce-file fname quiet)
    (let* ((file-as-list (tc:get-file-as-list fname))
	   ;; define-list is a list containing all define and grouped-define
	   ;; expressions in the file
	   (define-list (append
			 (apply append
				(map cdr
				     (tc:remove-pred
				      (lambda (x)
					(not (tc:grouped-defines? x)))
				      file-as-list)))
			 (tc:remove-pred
			  (lambda (x) (not (tc:define-exp? x)))
			  file-as-list)))
	   ;; deftype-list is a list containing all deftype, grouped-deftype,
	   ;; and define-record expressions in the file
	   (deftype-list (append
			  (apply append
				 (map cdr
				      (map tc:define-record->deftypes
					   (tc:remove-pred
					    (lambda (x)
					      (not (tc:define-record? x)))
					    file-as-list))))
			  (apply append
				 (map cdr
				      (tc:remove-pred
				       (lambda (x)
					 (not (tc:grouped-deftypes? x)))
				       file-as-list)))
			  (tc:remove-pred
			   (lambda (x) (not (tc:deftype? x)))
			   file-as-list)))
	   ;; load-list is a list of all load expression in the file
	   (load-list (tc:remove-pred
		       (lambda (x) (not (tc:load-like-exp? x)))
		       file-as-list))
	   ;; var-list is a list of the names being defined in the file
	   (var-list (map cadr define-list))
	   ;; value-list is a list of all the desugared bodies of the define
	   ;; expressions in the file
	   (value-list (map (tc:compose tc:desugar cddr) define-list))
	   ;; desugared-define-list is a list of define expressions with
	   ;; their bodies desugared
	   (desugared-define-list (tc:map2 (lambda (x y)
					     (list 'define x (car y)))
					   var-list value-list))
	   ;; exps-list is all the expressions in the file that were not
	   ;; put in one of the above lists
	   (exps-list (tc:remove-pred
		       (lambda (x) (or (tc:define-exp? x)
				       (tc:load-like-exp? x)
				       (tc:grouped-defines? x)
				       (tc:define-record? x)
				       (tc:deftype? x)
				       (tc:grouped-deftypes? x)))
		       file-as-list)))
      ;; do the loads
      (tc:process-loads load-list fname quiet)
      ;; do the deftypes
      (let ((perhaps-quietly-type-check
	     (if (not quiet)
		 type-check
		 (tc:type-check-maker
		  (lambda (name) (display ""))
		  (lambda (typ) (display ""))))))
	(for-each perhaps-quietly-type-check deftype-list))
      
      ;; do the defines
      ;; Create a type environment with the names of all defines letrec-
      ;; bound to new type variables.  Type check each define in the new
      ;; environment.  Add the new types to the global type environment.
      (letrec ((letrec-env
		;; TYPE: (-> ((list symbol)) (env symbol binding))
		(lambda (vl)
		  ;; INPUT: list of the names from the top-level define exps
		  ;; OUTPUT: a new type environment with the names letrec-
		  ;;         bound to new type variables
		  (if (null? vl)
		      (tc:env-empty)
		      (tc:env-update (car vl)
				     (tc:bind-type-expr
				      'letrec
				      (tc:new-variable-type-expr))
				     (letrec-env (cdr vl)))))))
	;; Create the mutual recursion environments
	(let ((letrec-bound-recur-env (letrec-env var-list)))
	  (set! tc:e-env (tc:env-empty))
	  (if (>= *tc:tracing-level* 2)
	      (displayln "*** letrec-bound-recur-env = "
			 letrec-bound-recur-env))
	  ;; Type check each top-level define expression using the mutual
	  ;; recursion environment, creaing a list of the types
	  (let* ((pass1-define-types
		  (map
		   (lambda (x)
		     ;; TYPE: (-> (define-expr) type-expr)
		     (tc:type-aux x letrec-bound-recur-env))
		   desugared-define-list))
		 (pass2-define-types
		  (map (lambda (x) (tc:subst-bindings tc:e-env x))
		       pass1-define-types)))
	    ;; Add the top-level define types to the global environment
	    (tc:map2 (lambda (x y) (tc:new-global x y quiet))
		     var-list
		     pass2-define-types))))
      
      ;; (displayln "about to do the expressions")
      ;; check each other expression
      (tc:check-each-expression exps-list quiet)
      )))

(define tc:new-global
  ;; TYPE: (-> (symbol type-expr boolean) void)
  (lambda (name its-type quiet)
    ;; MODIFIES: *tc:global-var-types*
    ;; EFFECT: install the new type, and print a message about it
    ;; Note: for consistency with type-check-maker,
    ;;       and to prevent cascading error messages, errors are installed too
    (set! *tc:global-var-types* 
	  (tc:env-update name its-type *tc:global-var-types*))
    (if (not quiet)
	(begin
	  (display name)
	  (display tc:colon)
	  (if (tc:error-type? its-type)
	      (displayln tc:cannot-be-typed)
	      (displayln (tc:type-translate its-type)))))))

(define tc:check-each-expression
  ;; TYPE: (-> ((list expression) boolean) void)
  (lambda (exps-list quiet)
    ;; REQUIRES: the expression is not a define or grouped define or deftype
    (for-each
     (lambda (x)
       (let ((res (tc:type (tc:desugar x))))
	 (if (tc:error-type? res)
	     (displayln tc:cannot-be-typed)
	     (if (not quiet)
		 (begin
		   (write x)
		   (display tc:colon)
		   (displayln (tc:type-translate res)))))))
     exps-list)))
 
(define tc:member-ls
  ;; TYPE: (-> ((list T) (list (pair T S))) (list (pair T S)))
  (lambda (name-ls def-ls)
    ((tc:flat-recur '() (lambda (x y) (if (tc:member? (car x) name-ls)
					  (cons x y)
					  y)))
     def-ls)))

(define tc:compose
  ;; TYPE: (-> ((-> (T) U) (-> (S) T)) (-> (S) U))
  (lambda (f g)
    (lambda (x)
      (f (g x)))))

(define tc:remove-pred
  ;; TYPE: (-> ((-> (T) boolean) (list T)) (list T))
  (lambda (pred ls)
    (cond
     ((null? ls) '())
     ((pred (car ls)) (tc:remove-pred pred (cdr ls)))
     (else (cons (car ls) (tc:remove-pred pred (cdr ls)))))))

(define tc:define-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (datum)
    (and (pair? datum) (eq? 'define (car datum)))))

(define tc:SICP-define-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (datum)
    (and (tc:define-exp? datum)	       
	 (>= 3 (length datum))
	 (pair? (cadr datum))
	 (symbol? (caadr datum)))))	; syntax error if we get here and
					; this isn't true

(define tc:load-like-exp?
  ;; TYPE: (-> (s-expression) boolean)
  (lambda (datum)
    (and (pair? datum)
	 (memq (car datum)
	       '(load load-from-lib load-quietly-from-lib load-quietly)))))

(define tc:is-def-file?
  ;; TYPE: (-> (string) boolean)
  (lambda (fname)
    (equal? (tc:get-file-suffix fname) "def")))

(define tc:is-scm-file?
  ;; TYPE: (-> (string) boolean)
  (lambda (fname)
    (equal? (tc:get-file-suffix fname) "scm")))

(define tc:get-file-prefix
  ;; TYPE: (-> (string) string)
  (lambda (name)
    (let ((dot-pos (tc:last-dot-in-file-pos name (- (string-length name) 1))))
      (if (= -1 dot-pos)
	  name
	  (substring name 0 dot-pos)))))

(define tc:get-file-suffix
  ;; TYPE: (-> (string) string)
  (lambda (name)
    (let* ((len (string-length name))
	   (dot-pos (tc:last-dot-in-file-pos name (- len 1))))
      (if (= -1 dot-pos)
	  ""
	  (substring name (+ 1 dot-pos) len)))))

(define tc:last-dot-in-file-pos
  ;; TYPE: (-> (string number) number)
  (lambda (name pos)
    (if (< pos 0)
	-1
	(let ((current-char (string-ref name pos)))
	  (if (char=? #\. current-char)
	      pos
	      (if (or (char=? #\/ current-char) (char=? #\\ current-char))
		  -1
		  (tc:last-dot-in-file-pos name (- pos 1))))))))

(define tc:handle-defs-file
  ;; TYPE: (-> (string boolean) void)
  (lambda (fname quiet)
    ;; EFFECT: places the contents of fname and the .def file
    ;;         associated with fname into lists.  It then performs
    ;;         the necessary
    ;;         type inference, unification, matching, and updating of
    ;;         *tc:global-var-types*.  Finally, it outputs the types
    ;;         for each procedure declared in fname, if this is not quiet.
    ;; (displayln "processing .def file for " fname)
    (tc:announce-file fname quiet)
    (let* ((file-as-list (tc:get-file-as-list fname))
	   ;; define-list is a list containing all define and grouped-define
	   ;; expressions in the .scm file
	   (define-list (append
			 (apply append
				(map cdr
				     (tc:remove-pred
				      (lambda (x)
					(not (tc:grouped-defines? x)))
				      file-as-list)))
			 (tc:remove-pred
			  (lambda (x) (not (tc:define-exp? x)))
			  file-as-list)))
	   ;; deftype-list is a list containing all deftype, grouped-deftype,
	   ;; and define-record expressions in the file
	   (ext-deftype-list (append
			      (apply append
				     (map cdr
					  (map tc:define-record->deftypes
					       (tc:remove-pred
						(lambda (x)
						  (not (tc:define-record? x)))
						file-as-list))))
			      (apply append
				     (map cdr
					  (tc:remove-pred
					   (lambda (x)
					     (not (tc:grouped-deftypes? x)))
					   file-as-list)))
			      (tc:remove-pred
			       (lambda (x) (not (tc:deftype? x)))
			       file-as-list)))
	   (deftype-list (map (lambda (dt)
				(list (car dt)
				      (cadr dt)
				      (tc:type-untranslate (caddr dt))))
			      ext-deftype-list))
	   ;; load-list is a list of all load expressions in the file
	   (load-list (tc:remove-pred
		       (lambda (x) (not (tc:load-like-exp? x)))
		       file-as-list))
	   ;; var-list is a list of the names being defined in the file
           (var-list (map cadr define-list))
	   ;; value-list is a list of all the desugared bodies of the define
	   ;; expressions in the file
           (value-list (map (tc:compose tc:desugar cddr) define-list))
	   ;; desugared-define-list is a list of define expressions with
	   ;; their bodies desugared
	   (desugared-define-list (tc:map2 (lambda (x y)
					     (list 'define x (car y)))
					   var-list value-list))
	   ;; exps-list is all the expressions in the file that were not put
	   ;; in one of the above lists
	   (exps-list (tc:remove-pred
		       (lambda (x) (or (tc:define-exp? x)
				       (tc:load-like-exp? x)
				       (tc:grouped-defines? x)
				       (tc:define-record? x)
				       (tc:deftype? x)
				       (tc:grouped-deftypes? x)))
		       file-as-list))
	   ;; the .def file is read below
	   (def-file-name (string-append (tc:get-file-prefix fname) ".def"))
	   (actual-def-file (tc:get-file-as-list def-file-name))
	   ;; check to see if the trustme! statement is included in the
	   ;; .def file
	   (trusted (and (not (null? actual-def-file))
			 (eq? (car actual-def-file) 'trustme!)))
	   ;; strip any leading trustme! statement from the .def file
	   (def-file (if trusted
			 (cdr actual-def-file)
			 actual-def-file))
	   ;; the-defreps is a list of pairs where the first element of
	   ;; each pair is an abstract type-expr and the second element
	   ;; is the corresponding representation type-expr.  It is
	   ;; generated from the defrep statements in the .def file.
	   (the-defreps (tc:get-reps def-file '()))
	   ;; type-def-ls is a list of pairs generated from the deftype
	   ;; statements in the .def file.  The first element of each pair
	   ;; is the symbol being bound in the deftype while the second element
	   ;; is the declared type of that symbol.
	   (type-def-ls
	    ;; TYPE: (list (pair symbol (pair type-expr null)))
	    (map (lambda (dft)
		   (if (and (list? dft)
			    (= 3 (length dft))
			    (eq? (car dft) 'deftype))
		       (list (cadr dft) (tc:type-untranslate (caddr dft)))
		       (error "bad syntax in file" def-file-name ": " dft)))
		 (tc:remove-pred (lambda (d) (eq? (car d) 'defrep))
				 def-file)))
	   ;; type-match? is a procedure for matching inferred and declared
	   ;; types based on a mapping from abstract types to representation
	   ;; types as defined in the defrep statements
	   (type-match? (tc:rep-type-match the-defreps))
	   ;; process-def-vs-inferred is called for each procedure defined
	   ;; in the .scm file.  It looks up the declared type of the
	   ;; procedure and performs matching vs. the inferred type.
	   (process-def-vs-inferred
	    ;; TYPE: (-> (symbol type-expr) void)
	    (lambda (name inferred-type)
	      ;; EFFECT: check the inferred type vs. the declared type,
	      ;;         and enter it in the globals if it matches.  If,
	      ;;         however, the inferred type was not declared, then
	      ;;         it is a local definition so its inferred
	      ;;         type is added to the local environment for use
	      ;;         in type inference for the rest of the .scm file.
	      (let* ((decl-assq (assq name type-def-ls))
		     (declared-type (if (not decl-assq)
					*tc:failure*
					(cadr decl-assq))))
		(if (not (tc:error-type? declared-type))
		    (let ((match-res (type-match? name
						  inferred-type
						  declared-type)))
		      (if (not (tc:error-type? match-res))
			  (tc:new-global name declared-type quiet)
			  (tc:new-global name *tc:failure* quiet)))
		    )))))		; end let* bindings
      ;; do the loads
      (tc:process-loads load-list fname quiet)
      ;; do the rest
      (if trusted
	  ;; If the .def file is trusted then the .scm file is not type-
	  ;; checked.  The bindings of symbols to declared types are simply
	  ;; entered into the global var types.
	  (begin
	    (for-each
	     (lambda (name-and-type)
	       (tc:new-global (car name-and-type) (cadr name-and-type) quiet))
	     type-def-ls))
	  ;; If the .def file is not trusted then the .scm file is type-
	  ;; checked and the inferred types and matched vs. the declared
	  ;; types.
	  (begin
	    ;; do the deftypes
	    ;; by matching deftypes in the .scm file with corresponding
	    ;; deftypes in the .def file.
;	    (displayln "about to do the deftypes")
	    (for-each
	     (lambda (dt)
	       (process-def-vs-inferred
		(cadr dt)
		(caddr dt)))
	     deftype-list)
;	    (displayln "about to do the defines")
	    ;; do the defines
	    ;; by inferring the type of each define expression in the .scm
	    ;; file and matching with the corresponding deftypes in the .def
	    ;; file
	    ;; Create a type environment with the names of all defines letrec-
	    ;; bound to new type variables.  Type check each define in the new
	    ;; environment.  Match the inferred types vs. the declared types.
	    (letrec
		((letrec-env
		  ;; TYPE: (-> ((list symbol)) (env symbol binding))
		  (lambda (vl)
		    ;; INPUT: list of the names from the top-level define exps
		    ;; OUTPUT: a new type environment with the names letrec-
		    ;;         bound to new type variables
		    (if (null? vl)
			(tc:env-empty)
			(tc:env-update (car vl)
				       (tc:bind-type-expr
					'letrec
					(tc:new-variable-type-expr))
				       (letrec-env (cdr vl))))))
		 (let-env
		  ;; TYPE: (-> ((list (pair symbol (pair symbol type-expr))))
		  ;;           (env symbol let-binding))
		  (lambda (dtl)
		    (if (null? dtl)
			(tc:env-empty)
			(tc:env-update (cadr (car dtl))
				       (tc:bind-type-expr
					'let
					(caddr (car dtl)))
				       (let-env (cdr dtl)))))))
			
	      ;; Create the mutual recursion environments
	      (let* ((local-define-env (letrec-env var-list))
		     (local-deftype-env (let-env deftype-list))
		     (letrec-bound-recur-env
		      (tc:env-join local-define-env local-deftype-env)))
		(set! tc:e-env (tc:env-empty))
		(if (>= *tc:tracing-level* 2)
		    (displayln "*** letrec-bound-recur-env = "
			       letrec-bound-recur-env))
		;; Type check each top-level define expression using the mutual
		;; recursion environment, creaing a list of the types
		(let* ((pass1-define-types
			(map
			 (lambda (x)
			   ;; TYPE: (-> (define-expr) type-expr)
			   (tc:type-aux x letrec-bound-recur-env))
			 desugared-define-list))
		       (pass2-define-types
			(map (lambda (x) (tc:subst-bindings tc:e-env x))
			     pass1-define-types))
		       (inferred-type-ls
			(tc:map2 (lambda (x y) (list x y))
				 var-list
				 pass2-define-types)))
		  (for-each
		   (lambda (name-and-type)
		     (process-def-vs-inferred (car name-and-type)
					      (cadr name-and-type)))
		   inferred-type-ls)
		  ;; output errors for names that have declared types but
		  ;; are not defined in the .scm file
		  (let ((inferred-names (append
					 (map cadr deftype-list)
					 var-list))
			(declared-names (map car type-def-ls)))
		    (for-each
		     (lambda (name)
		       (if (not (memq name inferred-names))
			   (tc:small-error
			    "Missing define in " fname " for declared name "
			    name)))
		     declared-names)))))
	    ;; (displayln "about to do the expressions")
	    ;; check each other expression
	    (tc:check-each-expression exps-list quiet))))))

(define tc:rep-type-match
  ;; TYPE: (-> ((list (pair symbol type-expr)))
  ;;           (-> (symbol type-expr type-expr) type-expr))
  (lambda (type-reps)
    (letrec ((mult-def-down-type
	      ;; TYPE: (-> ((list (pair type-expr type-expr)) type-expr)
	      ;;           type-expr)
	      (lambda (tr te)
		(if (null? tr)
		    te
		    (mult-def-down-type (cdr tr)
					(tc:down-type te (car tr)))))))
      (lambda (name inferred-type declared-type)
	;; EFFECT: match the inferred-type against the declared-type
	;;         if they unify, return the type,
	;;         otherwise issue an error and return *tc:failure*
	;;(displayln "tc:rep-type-match " name
	;;           " inferred-type " inferred-type
	;;           " declared-type " declared-type)
	(let* ((down-declared-type
		(mult-def-down-type type-reps declared-type))
	       (down-inferred-type
		(mult-def-down-type type-reps inferred-type))
	       (unif-res (tc:unify
			  down-inferred-type
			  down-declared-type
			  (tc:env-empty)
			  '())))
	  ;(displayln " down-inferred-type " down-inferred-type
	  ;           " down-declared-type " down-declared-type)
	  (if (not (tc:unification-error? unif-res))
	      declared-type
	      (begin
		(tc:small-report
		 "Definition does not match declared type (in .def file) for"
		 name
		 "Expected (declared) type"
		 (tc:type-translate down-declared-type)
		 "Inferred type           "
		 (tc:type-translate down-inferred-type))
		*tc:failure*)))))))

(define tc:get-reps
  ;; TYPE: (-> ((list datum) (list (pair type-expr type-expr)))
  ;;           (list (pair type-expr type-expr)))
  (lambda (ls tr-acc)
    ;; EFFECT: returns a list of pairs containing abstract type
    ;;         followed by the representation type.
    (cond
     ((null? ls)
      tr-acc)
     ((and (pair? (car ls)) (equal? (caar ls) 'defrep))
      (if (and (list? (car ls)) (= 3 (length (car ls))))
	  (tc:get-reps (cdr ls) (cons (cons (tc:type-untranslate (cadar ls))
					    (tc:type-untranslate (caddar ls)))
				      tr-acc))
	  (error "bad syntax for defrep in def file: " (car ls))))
     (else (tc:get-reps (cdr ls) tr-acc)))))

(define tc:get-file-as-list
  ;; TYPE: (-> (string) (list datum))
  (lambda (fname)
    (letrec ((get-file-aux
	      (lambda ()
		(let ((expr (read)))
		  (cond
		   ((eof-object? expr) '())
		   ((tc:define-exp? expr)
		    (cons (tc:handle-define expr)
			  (get-file-aux)))
		   (else (cons expr (get-file-aux))))))))
      (with-input-from-file fname get-file-aux))))

;;; Gary: The following is not used so I didn't modify it to use the
;;; type-expr ADT.   - Curt, 5/26/99
(define tc:subst-all
  ;; TYPE: (-> (type-expr type-expr (or type-expr (tree type-expr)))
  ;;           (or type-expr (tree type-expr)))
  (lambda (new old tr)
    ;; ENSURES: result is tr with all occurrences of old replaced by new
    (if (equal? tr old)
	new
	(if (symbol? tr)
	    tr
	    ((tc:deep-recur-abs
	      '()
	      (lambda (a d) (if (equal? a old) (cons new d) (cons a d)))
	      (lambda (a d) (if (equal? a old) (cons new d) (cons a d)))
	      (lambda (x) (or (symbol? x) (string? x))))
	     tr)))))


(define type-check-loop 
  ;; TYPE: (-> () void)
  (lambda ()
    (tc:type-check-loop-helper
     "enter expression (or done to exit): ")))

(define tc:type-check-loop-helper 
  ;; TYPE: (-> (string) void)
  (lambda (prompt-str)
    ;; EFFECT: prompt with prompt-str, read expressions and tc:type check,
    ;;         until done or EOF is entered
    (display prompt-str)
    (let ((expr (read)))
      (cond
       ((or (eof-object? expr) (eq? expr 'done))
	(display "Bye.")
	(newline))
       (else (type-check expr)
	     (tc:type-check-loop-helper prompt-str))))))


(define tc:desugar 
  ;; TYPE: (-> (s-expression) s-expression)
  (let ((lambda-body-desugar (tc:implicit->explicit "lambda body"))
	(let/rec-body-desugar (tc:implicit->explicit "let or letrec body"))
    	(let*-body-desugar (tc:implicit->explicit "let* body")))
  (lambda (exp)
    ;; REQUIRES: there are no syntax errors in exp
    ;; ENSURES: result is the same as the argument except that
    ;;          nested define is tc:replaced by letrec, cond by nested ifs.
    (if (>= *tc:tracing-level* 3)
	(displayln "entering tc:desugar with: " exp))
    (cond
     ((not (pair? exp)) exp)
     ((not (list? exp))
      (error "bad syntax: " exp))
     ;; now we know that exp is a non-empty list...
     ((eq? (car exp) 'quote)
      (if (not (= 2 (length exp)))
	  (error "wrong number of arguments for quote: " exp)
	  exp))
     ((eq? (car exp) 'quasiquote)
      (if (not (= 2 (length exp)))
	  (error "wrong number of arguments for quasiquote: " exp)
	  (tc:expand-quasiquote (cadr exp))))
     ;; Nested defines are desugared but top-level defines are not
     ((eq? (car exp) 'define)
      (if (< (length exp) 3)
	  (error "wrong number of arguments for define: " exp)
	  (let ((d (tc:desugar (cdr (tc:handle-define exp)))))
	    (cons 'define d))))
     ((eq? (car exp) 'cond)
      (tc:cond->if (cdr exp)))
     ((eq? (car exp) 'case)
      (list 'let
	    (list (list 'case-var!- (tc:desugar (cadr exp))))
	    (tc:case->if 'case-var!- (cddr exp))))
     ((eq? (car exp) 'lambda)
      (if (< (length exp) 3)
	  (error "wrong number of arguments for lambda: " exp)
	  (list 'lambda
		(cadr exp)
		(lambda-body-desugar (cddr exp)))))
     ((or (eq? (car exp) 'let)
	  (eq? (car exp) 'letrec))
      (if (< (length exp) 3)
	  (error
	   (string-append "wrong number of arguments for "
			  (symbol->string (car exp))
			  ": ")
	   exp)
	  (list (car exp)
		(map tc:desugar-decl (cadr exp))
		(let/rec-body-desugar (cddr exp)))))
     ((eq? (car exp) 'let*)
      (if (< (length exp) 3)
	  (error "wrong number of arguments for let*: " exp)
	  (cons 'let (let*->let (cadr exp)
				(let*-body-desugar (cddr exp))))))
     ((eq? (car exp) 'variant-case)
      (if (< (length exp) 3)
	  (error "wrong number of arguments for variant-case: " exp)
	  (tc:variant-case->if exp)))
     ((eq? (car exp) 'define-record)
      (if (not (= (length exp) 3))
	  (error "wrong number of arguments for define-record: " exp)
	  (tc:define-record->deftypes exp)))
     ;; [[[should handle "do" here.]]]
     (else ; applications
      (map tc:desugar exp))))))

(define tc:question-name
  ;; TYPE: (-> (symbol) symbol)
  (lambda (name)
    (string->symbol
     (string-append (symbol->string name) "?"))))

(define tc:field-access-name
  ;; TYPE: (-> (symbol) symbol)
  (lambda (record-name field-name)
    (string->symbol
     (string-append (symbol->string record-name)
		    "->"
		    (symbol->string field-name)))))

(define tc:make-record-name
  ;; TYPE: (-> (symbol) symbol)
  (lambda (record-name)
    (string->symbol
     (string-append "make-" (symbol->string record-name)))))

(define tc:variant-case->if
  ;; TYPE: (-> (s-exprssion) s-expression)
  (letrec ((clause-syntax-ok? ;; TYPE: (-> (datum) boolean)
	    (lambda (clause)
	      (and (list? clause)
		   (let ((len (length clause)))
		     (and (>= len 2)
			  (let ((tag (car clause)))
			    (and (symbol? tag)
				 (or (eq? 'else tag)
				     (= len 3)))))))))
	   (new-var '*record*-50014) ; hack, should really generate new name.  
	   (var-clause->cond-clause
	    ;; TYPE: (-> (list s-expression) (list s-expression))
	    (lambda (clause)
	      (if (not (clause-syntax-ok? clause))
		  (error "bad syntax in variant-case clause: " clause)
		  (let ((tag (car clause)))
		    (if (eq? 'else tag)
			clause
			(let ((field-names (cadr clause))
			      (body (cddr clause)))
			  (cond
			   ((eq? 'else tag) clause)
			   ((null? field-names)
			    (cons (list (tc:question-name tag) new-var) body))
			    (else
			      (list
			       (list (tc:question-name tag) new-var)
			       (append
				(list 'let
				      (map
				       (lambda (field-name)
					 (list field-name
					       (list 
						(tc:field-access-name
						 tag
						 field-name)
						 new-var)))
				       field-names))
				body)))))))))))
    (lambda (exp)
      (let ((tested-exp (car exp)))
	(list 'let (list (list new-var (tc:desugar (cadr exp))))
	      (tc:cond->if
	       (map var-clause->cond-clause (cddr exp))))))))

(define tc:nested-define->letrec
  ;; TYPE: (-> (s-expression) s-expression)
  (lambda (exp)
    ;; REQUIRES: exp is a syntactically-legal define expression
    ;;           of form (define name exp)
    (letrec ((no-defines
	      (lambda (exp acc)
		(cond 
		 ((null? exp) acc)
		 ((not (pair? exp)) acc) ; Should never reach this case..
		 ((define? (car exp)) 
		  (no-defines (cdr exp) acc))
		 (else (no-defines (cdr exp) (cons (car exp) acc))))))
             (get-defines 
	      (lambda (exp acc)
		(cond 
		 ((null? exp) acc)
		 ((not (pair? exp)) acc) ; Should never reach this case..
		 ((define? (car exp)) 
		  (get-defines (cdr exp) 
			       (cons (tc:sicp->sap (car exp))
				     acc)))
		 (else (get-defines (cdr exp) acc)))))
             (define? (lambda (exp) 
                        (and (pair? exp) 
                             (not (null? exp)) 
                             (eq? (car exp) 'define)))))
      (if (or (pair? (tc:second exp));; SICP style proc def
	      (tc:lambda-exp? (tc:third exp)));; just a straight def...
	  (let* ((name (cadr exp))
		 (args (car (cdaddr exp)))
		 (body (cdr (cdaddr exp)))
		 (letrec-exp (map cdr (get-defines body '())))
		 (new-body (reverse (no-defines body '()))))
	    (if (null? (get-defines body '()))
		exp
		(cons 'define 
		      (cons name (list (list 'lambda args 
					     (append 
					      (list 'letrec 
						    letrec-exp)
					      new-body)))))))
	  (list 'define (tc:second exp) (tc:third exp))))))

(define tc:desugar-decl
  ;; TYPE: (-> (decl) decl)
  (lambda (decl)
    (begin
      (if (not (= (length decl) 2))
	  (error "bad declaration syntax: " decl))
      (list (tc:first decl)
	    (tc:desugar (cadr decl))))))

(define let*->let
  ;; TYPE: (-> ((list decl) s-expression (list datum)))
  (lambda (decls desugared-body)
    (letrec ((decls-trans
	      ;; TYPE: (-> ((list decl) (list datum)))
	      (lambda (decls)
		(list (list (tc:desugar-decl (car decls)))
		      (if (tc:singleton-list? decls)
			  desugared-body
			  (cons 'let (decls-trans (cdr decls))))))))
      (if (null? decls)
	  (list decls desugared-body)
	  (decls-trans decls)))))


(define tc:sicp->sap
  ;; TYPE: (-> (s-expression) s-expression)
  (lambda (exp)
    ;; REQUIRES: exp is a define expression
    ;; EFFECT: changes a define that defines a procedure using the SICP syntax
    ;;         to the SAP syntax.
    ;;         ie, (define (name arg) ...) to (define name (lambda (arg) ...))
    ;; and pass through other define expressions (either SAP style or others)
    (cond 
     ((and (tc:define-exp? exp) (pair? (cadr exp))) ; SICP style define
      (let* ((name (caadr exp))
	     (arg-list (cdadr exp))
	     (body (cddr exp)))
	(list 'define name (append (list 'lambda arg-list) body))))
     (else exp))))

(define tc:handle-define (tc:compose tc:nested-define->letrec tc:sicp->sap))

(define tc:cond->if
  ;; TYPE: (-> ((list (list s-expression))) s-expression)
  (let ((true-part
	 ;; TYPE: (-> ((list s-expression)) s-expression)
	 (tc:implicit->explicit "cond clause")))
    (lambda (clauses)
      ;; REQUIRES: there are no syntax errors in the clauses
      ;; ENSURES: result is a desugared if-expression equivalent to the cond
      (cond
       ((null? clauses)
	(error "cond must contain at least one clause: " clauses))
       ((not (and (pair? clauses)
		  (pair? (car clauses))))
	(error "bad syntax in cond expression, expecting a clause: " clauses))
       ((and (null? (cdr clauses)) 
	     (eq? (caar clauses) 'else))
	(true-part (cdar clauses)))
       ((null? (cdr clauses))
	(if *allow-cond-expressions-without-else-clause* 
	    (list 'if
		  (tc:desugar (caar clauses))
		  (true-part (cdar clauses)))
	    (error "cond must have an else clause: " clauses)))
       (else (list 'if
		   (tc:desugar (caar clauses))
		   (true-part (cdar clauses))
		   (tc:cond->if (cdr clauses))))))))

(define tc:case->if
  ;; TYPE: (-> ((list (list s-expression))) s-expression)
  (let ((true-part
	 ;; TYPE: (-> ((list s-expression)) s-expression)
	 (tc:implicit->explicit "case clause")))
    (lambda (tested-exp clauses)
      ;; REQUIRES: there are no syntax errors in the clauses
      ;; ENSURES: result is an if expression equivalent to the case clauses
      ;; NOTE: this is *not* dynamically correct unless tested-exp is constant
      (cond
       ((null? clauses)
	(error "case must contain at least one clause: " clauses))
       ((not (and (pair? clauses)
		  (pair? (car clauses))))
	(error "bad syntax in case expression, expecting a clause: " clauses))
       ((and (null? (cdr clauses)) 
	     (eq? (caar clauses) 'else))
	(true-part (cdar clauses)))
       ((null? (cdr clauses))
	(if *allow-case-expressions-without-else-clause* 
	    (list 'if
		  (list 'memv tested-exp (list 'quote (caar clauses)))
		  (true-part (cdar clauses)))
	    (error "case must have an else clause: " clauses)))
       (else (list 'if
		   (list 'memv tested-exp (list 'quote (caar clauses)))
		   (true-part (cdar clauses))
		   (tc:case->if tested-exp (cdr clauses))))))))


(define tc:type-write
  ;; TYPE: (-> (external-type-expr) void)
  display)

(define tc:expand-quasiquote
  ;; TYPE: (-> (s-expression) s-expression)
  (lambda (template)
    ;; ENSURES: expands the quasiquotation template,
    ;; taking unquote and unquote-splicing into account
    ;; NOTE: this assumes that cons and append are not redefined by the user!
    (cond
     ((or (null? template) (symbol? template))
      (list 'quote template))
     ((or (number? template) (char? template) (string? template)
	  (boolean? template))
      template)
     ((vector? template)
      (cons 'vector (map tc:expand-quasiquote (vector->list template))))
     ((pair? template)
      (cond
       ((and (eq? 'unquote (car template)))
	(if (and (list? template) (= 2 (length template)))
	    (tc:desugar (cadr template))
	    (error "Bad syntax in unquote (,): " template)))
       ((and (pair? (car template))
	     (eq? 'unquote-splicing (caar template)))
	(if (and (list? (car template)) (= 2 (length (car template))))
	    (list 'append
		  (tc:desugar (cadr (car template)))
		  (tc:expand-quasiquote (cdr template)))
	    (error "Bad syntax in unquote-splicing (,@): " (car template))))
       (else
	(list 'cons
	      (tc:expand-quasiquote (car template))
	      (tc:expand-quasiquote (cdr template)))))))))

(define tc:grouped-defines?
  ;; (-> (datum) boolean)
  (lambda (x)
    (and (list? x)
	 (> (length x) 1)
	 (eq? (car x) 'begin)
	 (tc:all tc:define-exp? (cdr x)))))

(define tc:grouped-deftypes?
  ;; (-> (datum) boolean)
  (lambda (x)
    (and (list? x)
	 (> (length x) 1)
	 (eq? (car x) 'begin)
	 (tc:all tc:deftype? (cdr x)))))

(define tc:define-record->deftypes
  ;; TYPE: (-> (s-expression) s-expression)
  (lambda (expr)
    ;; REQUIRES: expr looks like a define-record expression
    ;; ENSURES: result is a grouped deftype expression for expr
    ;; Note: deftype uses the external type syntax,
    ;;       so don't use the stuff in type-check-type-expr!
    (let ((record-name (cadr expr))
	  (field-names (caddr expr)))
      (if (not (tc:all symbol? field-names))
	  (error "bad syntax in define-record fields: " expr)
	  (let* ((field-type-vars (map (lambda (field-name)
					 (tc:new-variable))
				       field-names))
		 (record-type (if (null? field-names)
				  record-name
				  (cons record-name field-type-vars)))
		 (field-type-mapping (tc:map2 cons
					      field-names
					      field-type-vars)))
	    (cons
	     'begin
	     (append
	      (list
	       (list 'deftype
		     (tc:question-name record-name)
		     '(-> (datum) boolean))
	       (list 'deftype
		     (tc:make-record-name record-name)
		     (list '-> field-type-vars record-type)))
	      (map (lambda (field-name)
		     (list 'deftype
			   (tc:field-access-name record-name field-name)
			   (list '->
				 (list record-type)
				 (cdr (assq field-name field-type-mapping)))))
		   field-names))))))))
